/*
 * Copyright (c) 2015, 2016, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

package com.oracle.iot.sample.ext;

import com.oracle.iot.client.device.DirectlyConnectedDevice;
import com.oracle.iot.client.device.util.RequestDispatcher;
import com.oracle.iot.client.device.util.RequestHandler;
import com.oracle.iot.client.message.DataMessage;
import com.oracle.iot.client.message.RequestMessage;
import com.oracle.iot.client.message.ResponseMessage;
import com.oracle.iot.client.message.StatusCode;
import com.oracle.iot.client.message.AlertMessage;

import com.oracle.iot.sample.HumiditySensor;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.UUID;

/*
 * This sample presents a simple humiditySensor to the IoT server.
 *
 * It uses the MessageDispatcher utility to send messages and to handle
 * resource requests from the server.
 *
 * The simple humiditySensor is polled every half second and a data message is
 * sent to the server and an alert message if the when alert condition is met.
 *
 * Note that the code is Java SE 1.5 compatible.
 */
public class DirectlyConnectedDeviceSample {

    private static final String HUMIDITY_SENSOR_MODEL_URN = "urn:com:oracle:iot:device:humidity_sensor";
    private static final String HUMIDITY_ATTRIBUTE = "humidity";
    private static final String MAX_THRESHOLD_ATTRIBUTE = "maxThreshold";
    private static final String TOO_HUMID_ALERT = HUMIDITY_SENSOR_MODEL_URN + ":too_humid";

    /**
     * humiditySensor polling interval before sending next readings.
     * Could be configured using {@code com.oracle.iot.sample.sensor_polling_interval} property
     */
    private static final long SENSOR_POLLING_INTERVAL = 
        Long.getLong("com.oracle.iot.sample.sensor_polling_interval", 5000);

    private static final boolean useLongPolling =
        !Boolean.getBoolean("com.oracle.iot.client.disable_long_polling");

    public static boolean isUnderFramework = false;
    public static boolean exiting = false;

    public static void main(String[] args) {

        final HumiditySensor humiditySensor =
                new HumiditySensor(UUID.randomUUID().toString());

        DirectlyConnectedDevice directlyConnectedDevice = null;

        try {
            if (args.length < 2) {
                throw new IllegalArgumentException("\nIncorrect number of arguments.");
            }

            // Initialize device client
            directlyConnectedDevice = new DirectlyConnectedDevice(args[0], args[1]);

            // Activate the device
            if (!directlyConnectedDevice.isActivated()) {
                display("\nActivating...");
                directlyConnectedDevice.activate(HUMIDITY_SENSOR_MODEL_URN);
            }

            // Create the directly-connected device instance
            display("\nCreating the directly-connected device instance " +
                    directlyConnectedDevice.getEndpointId());

            // Register handler for attributes of the device model. The humidity humiditySensor
            // device model only has maxThreshold for a writable attribute.
            final RequestDispatcher requestDispatcher = RequestDispatcher.getInstance();
            requestDispatcher.registerRequestHandler(directlyConnectedDevice.getEndpointId(),
                "deviceModels/" + HUMIDITY_SENSOR_MODEL_URN + "/attributes",
                new RequestHandler() {
                    public ResponseMessage handleRequest(RequestMessage requestMessage) {
                        return handleHumidityAttributeRequest(requestMessage, humiditySensor);
                    }
                }
            );

            // The previous maximum humidity threshold.
            // If the new maximum threshold is not equal to the previous,
            // the new value will be sent in the data message.
            int prevMaxHumidityThreshold = humiditySensor.getMaxThreshold();

            display("\n\tPress enter to exit.\n");

            // a flag to make sure alerts are only sent when crossing the threshold
            boolean alerted = false;

            // Main loop for single-threaded example.
            for (; ; ) {

                final int humidity = humiditySensor.getHumidity();

                // Send data from the indirectly-connected device
                final DataMessage.Builder dataMessageBuilder =
                        new DataMessage.Builder();

                StringBuilder consoleMessage =
                        new StringBuilder(new Date().toString());

                dataMessageBuilder.format(HUMIDITY_SENSOR_MODEL_URN +
                        ":attributes")
                        .source(directlyConnectedDevice.getEndpointId())
                        .dataItem(HUMIDITY_ATTRIBUTE, humidity);
                consoleMessage.append(" : ")
                        .append(directlyConnectedDevice.getEndpointId())
                        .append(" : Data : \"")
                        .append(HUMIDITY_ATTRIBUTE)
                        .append("\"=")
                        .append(humidity);

                // Send humidityThreshold if it has changed
                int humidityThreshold = humiditySensor.getMaxThreshold();
                if (prevMaxHumidityThreshold != humidityThreshold) {
                    prevMaxHumidityThreshold = humidityThreshold;
                    dataMessageBuilder.dataItem(MAX_THRESHOLD_ATTRIBUTE, humidityThreshold);
                    consoleMessage.append(",\"")
                            .append(MAX_THRESHOLD_ATTRIBUTE)
                            .append("\"=")
                            .append(humidityThreshold);
                }

                DataMessage dataMessage = dataMessageBuilder.build();

                display(consoleMessage.toString());
                directlyConnectedDevice.send(dataMessage);

                if (humidity > humidityThreshold) {
                    if (!alerted) {
                        try {
                            alerted = true;
                            final AlertMessage alertMessage =
                                    new AlertMessage.Builder()
                                            .format(TOO_HUMID_ALERT).source(directlyConnectedDevice.getEndpointId())
                                            .description("max threshold crossed")
                                            .dataItem(HUMIDITY_ATTRIBUTE, humidity)
                                            .severity(AlertMessage.Severity.CRITICAL).build();

                            consoleMessage = new StringBuilder(new Date().toString())
                                    .append(" : ")
                                    .append(directlyConnectedDevice.getEndpointId())
                                    .append(" : Alert : \"humidity\"=")
                                    .append(humidity)
                                    .append(",\"maxThreshold\"=")
                                    .append(humidityThreshold);

                            display(consoleMessage.toString());
                            directlyConnectedDevice.send(alertMessage);
                        } catch (Exception e) {
                            System.err.println("could not raise alert: " + e.getMessage());
                        }
                    }
                }

                // Receive and dispatch requests from the server
                //
                // This loop is constructed so that it will Wait SENSOR_POLLING_INTERVAL
                // seconds before sending the next reading. If there is a request message,
                // the message is dispatched to the handler and the response message is sent.
                //
                //
                long receiveTimeout = SENSOR_POLLING_INTERVAL;
                while (receiveTimeout > 0) {

                    if (!isUnderFramework) {
                        // User pressed the enter key while sleeping, exit.
                        if (System.in.available() > 0) {
                            exiting = true;
                            break;
                        }
                    }

                    long before = System.currentTimeMillis();
                    RequestMessage requestMessage = directlyConnectedDevice.receive(receiveTimeout);

                    if (requestMessage != null) {
                        ResponseMessage responseMessage =
                                requestDispatcher.dispatch(requestMessage);
                        directlyConnectedDevice.send(responseMessage);

                        // return to main loop so that the updated values can be
                        // sent to the server.
                        break;
                    }

                    /*
                     * The receive call and the processing of a server request
                     * may take less than the device polling interval so reduce
                     * the receive timeout by the time taken.
                     */
                    receiveTimeout -= System.currentTimeMillis() - before;

                    if (!useLongPolling && receiveTimeout > 0) {
                        Thread.sleep(receiveTimeout);
                        break;
                    }
                }

                if (exiting) {
                    break;
                }
            }
        } catch (Throwable e) {
            // could be java.lang.NoClassDefFoundError
            displayException(e);
            if (isUnderFramework) {
                throw new RuntimeException(e);
            }
        } finally {
            // Dispose of the device client
            if (directlyConnectedDevice != null) {
                try {
                    directlyConnectedDevice.close();
                } catch (Exception ignored) {
                }
            }
        }
    }

    // Handle deviceModels/urn:com:oracle:iot:device:humidity_sensor/attributes[/maxThreshold]
    private static ResponseMessage handleHumidityAttributeRequest(RequestMessage requestMessage,
                                                                  HumiditySensor humiditySensor) {

        try {

            // The request body of a PUT looks like {"value":70}, and the attribute is suffixed to the path
            // The request body of a POST looks like {"maxThreshold":70}
            final String jsonRequestBody = new String(requestMessage.getBody(), "UTF-8");
            final JSONObject jsonObject = new JSONObject(jsonRequestBody);

            int value = jsonObject.has("value") ? jsonObject.getInt("value") : jsonObject.getInt("maxThreshold");
            humiditySensor.setMaxThreshold(value);

            final StringBuilder consoleMessage = new StringBuilder(new Date().toString())
                    .append(" : ")
                    .append(requestMessage.getDestination())
                    .append(" : Request : ")
                    .append("\"maxThreshold\"=")
                    .append(value);

            display(consoleMessage.toString());

            return new ResponseMessage.Builder(requestMessage)
                    .statusCode(StatusCode.OK)
                    .build();

        } catch (UnsupportedEncodingException ue) {
            // UTF-8 is required, so this won't happen,
            // but throw an exception to make the compiler happy.
            throw new RuntimeException(ue);
        } catch (JSONException je) {
            je.printStackTrace();
            return new ResponseMessage.Builder(requestMessage)
                    .statusCode(StatusCode.INTERNAL_SERVER_ERROR)
                    .build();
        }
    }

    private static void showUsage() {
        Class<?> thisClass = new Object() { }.getClass().getEnclosingClass();
        display("Usage:\n"
                + "java " + thisClass.getName()
                + " <trusted assets file> <trusted assets password>\n");
    }

    private static void display(String string) {
        System.out.println(string);
    }

    private static void displayException(Throwable e) {
        StringBuffer sb = new StringBuffer(e.getMessage() == null ? 
                  e.getClass().getName() : e.getMessage());
        if (e.getCause() != null) {
            sb.append(".\n\tCaused by: ");
            sb.append(e.getCause());
        }
        System.out.println('\n' + sb.toString() + '\n');
        showUsage();
    }
}
