/*
 * Copyright (c) 2015, 2016, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

package com.oracle.iot.sample;

import oracle.iot.client.DeviceModel;
import oracle.iot.client.device.Alert;
import oracle.iot.client.device.DirectlyConnectedDevice;
import oracle.iot.client.device.VirtualDevice;

import java.io.IOException;
import java.util.Date;
import java.util.UUID;

/*
 * This sample presents a simple sensor as virtual
 * device to the IoT server.
 *
 * The simple sensor is polled every half second and the virtual device is
 * updated. An alert is generated when the value from the sensor exceeds
 * the maximum threshold of the sensor's data model.
 *
 * Note that the code is Java SE 1.5 compatible.
 */
public class DirectlyConnectedDeviceSample {

    private static final String HUMIDITY_SENSOR_MODEL_URN =
        "urn:com:oracle:iot:device:humidity_sensor";
    private static final String HUMIDITY_ATTRIBUTE = "humidity";
    private static final String MAX_THRESHOLD_ATTRIBUTE = "maxThreshold";
    private static final String TOO_HUMID_ALERT =
        HUMIDITY_SENSOR_MODEL_URN + ":too_humid";
    /**
     * sensor polling interval before sending next readings.
     * Could be configured using {@code com.oracle.iot.sample.sensor_polling_interval} property
     */
    private static final long SENSOR_POLLING_INTERVAL = 
        Long.getLong("com.oracle.iot.sample.sensor_polling_interval", 5000);

    // The following calculations of number_of_loops and sleepTime break the
    // SENSOR_POLLING_INTERVAL into a number of smaller intervals approximately
    // SLEEP_TIME milliseconds long. This just makes the sample a little more
    // responsive to keyboard input and has nothing to do with the client library.
    private static final int SLEEP_TIME = 100;
    private static long number_of_loops = (SLEEP_TIME > SENSOR_POLLING_INTERVAL ? 1 :
                                           SENSOR_POLLING_INTERVAL / SLEEP_TIME); 
    private static long sleepTime = 
        (SLEEP_TIME > SENSOR_POLLING_INTERVAL ?
         SENSOR_POLLING_INTERVAL :
         SLEEP_TIME + (SENSOR_POLLING_INTERVAL - number_of_loops * SLEEP_TIME) / number_of_loops);          

    public static boolean isUnderFramework = false;
    public static boolean exiting = false;

    public static void main(String[] args) {



        DirectlyConnectedDevice directlyConnectedDevice = null;

        try {
            if (args.length != 2) {
                display("\nIncorrect number of arguments.\n");
                throw new IllegalArgumentException("");
            }
            
            // Initialize the device client
            directlyConnectedDevice = new DirectlyConnectedDevice(args[0], args[1]);

            // Activate the device
            if (!directlyConnectedDevice.isActivated()) {
                display("\nActivating...");
                directlyConnectedDevice.activate(HUMIDITY_SENSOR_MODEL_URN);
            }

            // The "real" device
            final HumiditySensor sensor =
                    new HumiditySensor(directlyConnectedDevice.getEndpointId() + "_Sample_HS");

            // Create a virtual device implementing the device model
            final DeviceModel deviceModel =
                directlyConnectedDevice.getDeviceModel(HUMIDITY_SENSOR_MODEL_URN);

            final VirtualDevice virtualHumiditySensor =
                directlyConnectedDevice.createVirtualDevice(directlyConnectedDevice.getEndpointId(), deviceModel);

            final Alert tooHumidAlert = virtualHumiditySensor.createAlert(TOO_HUMID_ALERT);

            // Initialize the sensor to the device model's maxThreshold default value
            int defaultThreshold = virtualHumiditySensor.get("maxThreshold");
            sensor.setMaxThreshold(defaultThreshold);
            // The previous maximum humidity threshold.
            // If the new maximum humidity threshold is not equal to the previous,
            // update "maxThreshold" in the virtual HumiditySensor
            int prevMaxHumidityThreshold = sensor.getMaxThreshold();

            /*
             * Monitor the virtual device for requested attribute changes and
             * errors.
             *
             * Since there is only one attribute, maxThreshold, this could have
             * done with using an attribute specific on change handler.
             */
            virtualHumiditySensor.setOnChange(
                new VirtualDevice.ChangeCallback<VirtualDevice>() {
                    public void onChange(
                            VirtualDevice.ChangeEvent<VirtualDevice> event) {

                        VirtualDevice virtualDevice =
                            event.getVirtualDevice();
                        VirtualDevice.NamedValue<?> namedValues =
                            event.getNamedValue();
                        StringBuilder msg =
                            new StringBuilder(new Date().toString());
                        msg.append(" : ");
                        msg.append(virtualDevice.getEndpointId());
                        msg.append(" : onChange : ");

                        boolean first = true;
                        for (VirtualDevice.NamedValue<?> namedValue =
                             namedValues;
                             namedValue != null;
                             namedValue = namedValue.next()) {
                            final String attribute = namedValue.getName();
                            final Object value = namedValue.getValue();
                            if (!first) {
                                msg.append(',');
                            } else {
                                first = false;
                            }

                            if (MAX_THRESHOLD_ATTRIBUTE.equals(
                                    attribute)) {
                                int max = ((Integer)value).intValue();
                                msg.append("\"maxThreshold\"=");
                                msg.append(max);

                                // Update the actual device
                                sensor.setMaxThreshold(max);
                                /*
                                 * After this callback returns without an
                                 * exception, the virtual device attribute
                                 * will be updated locally and on the server.
                                 */
                            } else {
                                msg.append("\"")
                                        .append(attribute)
                                        .append("\" not implemented");
                            }
                        }

                        display(msg.toString());
                    }
                });

            virtualHumiditySensor.setOnError(
                new VirtualDevice.ErrorCallback<VirtualDevice>() {
                    public void onError(VirtualDevice.ErrorEvent<VirtualDevice
                            > event) {
                        VirtualDevice device =  event.getVirtualDevice();
                        display(new Date().toString() + " : onError : " +
                                device.getEndpointId() +
                                " : \"" + event.getMessage() + "\"");
                    }
                });



            display("\nCreated virtual humidity sensor " + virtualHumiditySensor.getEndpointId());
            display("\n\tPress enter to exit.\n");

            /*
             * A flag to make sure alerts are only sent when
             * crossing the threshold.
             */
            boolean humidityAlerted = false;

            mainLoop:
            for (;;) {

                int humidity = sensor.getHumidity();

                StringBuilder consoleMessage = new StringBuilder(new Date().toString())
                        .append(" : ")
                        .append(virtualHumiditySensor.getEndpointId())
                        .append(" : Set : \"")
                        .append(HUMIDITY_ATTRIBUTE)
                        .append("\"=")
                        .append(humidity);

                // Set the virtualilzed humidity sensor value in an 'update'. If
                // the humidity maxThreshold has changed, that maxThreshold attribute
                // will be added to the update (see the block below). A call to
                // 'finish' will commit the update.
                virtualHumiditySensor.update()
                    .set(HUMIDITY_ATTRIBUTE, humidity);

                // Send humidityThreshold if it has changed or is not the default
                int humidityThreshold = sensor.getMaxThreshold();
                if (prevMaxHumidityThreshold != humidityThreshold) {
                    prevMaxHumidityThreshold = humidityThreshold;

                    consoleMessage
                            .append(",\"")
                            .append(MAX_THRESHOLD_ATTRIBUTE)
                            .append("\"=")
                            .append(humidityThreshold);

                    virtualHumiditySensor.set(MAX_THRESHOLD_ATTRIBUTE, humidityThreshold);
                }

                display(consoleMessage.toString());
                virtualHumiditySensor.finish();

                if (humidity > humidityThreshold) {
                    if (!humidityAlerted) {
                        humidityAlerted = true;
                        consoleMessage = new StringBuilder(new Date().toString())
                                .append(" : ")
                                .append(virtualHumiditySensor.getEndpointId())
                                .append(" : Alert : \"")
                                .append(HUMIDITY_ATTRIBUTE)
                                .append("\"=")
                                .append(humidity)
                                .append(",\"")
                                .append(MAX_THRESHOLD_ATTRIBUTE)
                                .append("\"=")
                                .append(humidityThreshold);
                        display(consoleMessage.toString());
                        tooHumidAlert.set(HUMIDITY_ATTRIBUTE, humidity).raise();
                    }
                } else {
                    humidityAlerted = false;
                }

                //
                // Wait SENSOR_POLLING_INTERVAL seconds before sending the next reading.
                // The SENSOR_POLLING_INTERVAL is broken into a number of smaller increments
                // to make the application more responsive to a keypress, which will cause
                // the application to exit.
                //
                for (int i = 0; i < number_of_loops; i++) {
                    Thread.sleep(sleepTime);

                    // when running under framework, use platform specific exit
                    if (!isUnderFramework) {
                        // User pressed the enter key while sleeping, exit.
                        if (System.in.available() > 0) {
                            break mainLoop;
                        }
                    } else if (exiting) {
                        break mainLoop;
                    }
                }
            }
        } catch (Throwable e) { 
            // catching Throwable, not Exception:
            // could be java.lang.NoClassDefFoundError
            // which is not Exception

            displayException(e);
            if (isUnderFramework) throw new RuntimeException(e);
        } finally {
            // Dispose of the device client
            try {
                if (directlyConnectedDevice != null) directlyConnectedDevice.close();
            } catch (IOException ignored) {
            }
        }
    }
    
    private static void showUsage() {
        Class<?> thisClass = new Object() { }.getClass().getEnclosingClass();
        display("Usage: \n"
                + "java " + thisClass.getName()
                + " <trusted assets file> <trusted assets password>\n"
        );
    }

    private static void display(String string) {
        System.out.println(string);
    }

    private static void displayException(Throwable e) {
        StringBuilder sb = new StringBuilder(e.getMessage() == null ?
                  e.getClass().getName() : e.getMessage());
        if (e.getCause() != null) {
            sb.append(".\n\tCaused by: ");
            sb.append(e.getCause());
        }
        System.out.println('\n' + sb.toString() + '\n');
        showUsage();
    }
}
