/*
 * Copyright (c) 2015, 2016, Oracle and/or its affiliates. All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL). See the LICENSE file in the root
 * directory for license terms. You may choose either license, or both.
 */

package com.oracle.iot.sample;

import oracle.iot.client.DeviceModel;
import oracle.iot.client.device.Alert;
import oracle.iot.client.device.GatewayDevice;
import oracle.iot.client.device.VirtualDevice;

import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import java.security.GeneralSecurityException;

/*
 * This sample is a gateway that presents multiple simple sensors as virtual
 * devices to the IoT server.
 *
 * The simple sensors are polled every half second and each virtual device
 * is updated. An alert is generated when the value from either sensor exceeds
 * the maximum threshold of that sensor's data model.
 *
 * Note that the code is Java SE 1.5 compatible.
 */
public class GatewayDeviceSample {

    private static final String HUMIDITY_SENSOR_MODEL_URN =
        "urn:com:oracle:iot:device:humidity_sensor";
    private static final String HUMIDITY_ATTRIBUTE = "humidity";
    private static final String MAX_THRESHOLD_ATTRIBUTE = "maxThreshold";
    private static final String TOO_HUMID_ALERT =
        HUMIDITY_SENSOR_MODEL_URN + ":too_humid";

    private static final String TEMPERATURE_SENSOR_MODEL_URN =
        "urn:com:oracle:iot:device:temperature_sensor";
    private static final String UNIT_ATTRIBUTE = "unit";
    private static final String TEMPERATURE_ATTRIBUTE = "temp";
    private static final String MIN_THRESHOLD_ATTRIBUTE = "minThreshold";
    private static final String MIN_TEMP_ATTRIBUTE = "minTemp";
    private static final String MAX_TEMP_ATTRIBUTE = "maxTemp";
    private static final String START_TIME_ATTRIBUTE = "startTime";
    private static final String TOO_COLD_ALERT =
        TEMPERATURE_SENSOR_MODEL_URN + ":too_cold";
    private static final String TOO_HOT_ALERT =
        TEMPERATURE_SENSOR_MODEL_URN + ":too_hot";
    /**
     * sensor polling interval before sending next readings.
     * Could be configured using {@code com.oracle.iot.sample.sensor_polling_interval} property
     */
    private static final long SENSOR_POLLING_INTERVAL =
        Long.getLong("com.oracle.iot.sample.sensor_polling_interval", 5000);

    // The following calculations of number_of_loops and sleepTime break the
    // SENSOR_POLLING_INTERVAL into a number of smaller intervals approximately
    // SLEEP_TIME milliseconds long. This just makes the sample a little more
    // responsive to keyboard input and has nothing to do with the client library.
    private static final int SLEEP_TIME = 100;
    private static long number_of_loops = (SLEEP_TIME > SENSOR_POLLING_INTERVAL ? 1 :
                                           SENSOR_POLLING_INTERVAL / SLEEP_TIME);
    private static long sleepTime =
        (SLEEP_TIME > SENSOR_POLLING_INTERVAL ?
         SENSOR_POLLING_INTERVAL :
         SLEEP_TIME + (SENSOR_POLLING_INTERVAL - number_of_loops * SLEEP_TIME) / number_of_loops);

    public static boolean isUnderFramework = false;
    public static boolean exiting = false;

    public static void main(String[] args) {

        GatewayDevice gatewayDevice = null;

        try {
            if (args.length < 2) {
                display("\nIncorrect number of arguments.");
                throw new IllegalArgumentException("");
            }

            display("\nCreating the gateway instance...");
            // Initialize the device client
            gatewayDevice = new GatewayDevice(args[0], args[1]);

            // Activate the device
            if (!gatewayDevice.isActivated()) {
                display("\nActivating...");
                gatewayDevice.activate();
            }

            //
            // The hardware id for the sensor instances are fabricated
            // from the gateway's endpoint id and a suffix, the combination of
            // which is likely to be unique. On subsequent runs of the
            // GatewayDeviceSample (with the same gateway device), the
            // endpoint id from the previously registered indirectly connected
            // device will be returned by the registerDevice call. In other
            // words, the same indirectly connected device will be used.
            //
            final TemperatureSensor temperatureSensor =
                    new TemperatureSensor(
                            args.length > 2
                                    ? args[2]
                                    : gatewayDevice.getEndpointId() + "_Sample_TS");
            final String temperatureSensorEndpointId;

            final HumiditySensor humiditySensor =
                new HumiditySensor(
                        args.length > 3
                                ? args[3]
                                : gatewayDevice.getEndpointId() + "_Sample_HS");

            Map<String, String> metaData = new HashMap<String, String>();
            metaData.put(
                         GatewayDevice.MANUFACTURER,
                         temperatureSensor.getManufacturer());
            metaData.put(
                         GatewayDevice.MODEL_NUMBER,
                         temperatureSensor.getModelNumber());
            metaData.put(
                         GatewayDevice.SERIAL_NUMBER,
                         temperatureSensor.getSerialNumber());

            //
            // If the user gave a hardware id for the temperature sensor,
            // then restrict the sensor to this gateway. This means that
            // the sensor cannot be connected through other gateways.
            //
            final boolean temperatureSensorRestricted = args.length > 2;
            temperatureSensorEndpointId =
                gatewayDevice.registerDevice(
                        temperatureSensorRestricted,
                        temperatureSensor.getHardwareId(),
                        metaData,
                        TEMPERATURE_SENSOR_MODEL_URN);

            metaData = new HashMap<String, String>();
            metaData.put(
                         GatewayDevice.MANUFACTURER,
                         humiditySensor.getManufacturer());
            metaData.put(
                         GatewayDevice.MODEL_NUMBER,
                         humiditySensor.getModelNumber());
            metaData.put(
                         GatewayDevice.SERIAL_NUMBER,
                         humiditySensor.getSerialNumber());

            //
            // If the user gave a hardware id for the humidity sensor,
            // then restrict the sensor to this gateways. This means that
            // the sensor cannot be connected through other gateways.
            //
            final boolean humiditySensorRestricted = args.length > 3;
            final String humiditySensorEndpointId =
                        gatewayDevice.registerDevice(
                                humiditySensorRestricted,
                                humiditySensor.getHardwareId(),
                                metaData,
                                HUMIDITY_SENSOR_MODEL_URN);

            // Create the virtual devices implementing the device models
            final DeviceModel temperatureDeviceModel =
                    gatewayDevice.getDeviceModel(TEMPERATURE_SENSOR_MODEL_URN);

            final VirtualDevice virtualTemperatureSensor =
                    gatewayDevice.createVirtualDevice(
                            temperatureSensorEndpointId,
                            temperatureDeviceModel);

            // Initialize the temperature sensor to the device model's min and maxThreshold default values
            int defaultThreshold = virtualTemperatureSensor.get("minThreshold");
            temperatureSensor.setMinThreshold(defaultThreshold);
            defaultThreshold = virtualTemperatureSensor.get("maxThreshold");
            temperatureSensor.setMaxThreshold(defaultThreshold);

            final DeviceModel humidityDeviceModel =
                    gatewayDevice.getDeviceModel(HUMIDITY_SENSOR_MODEL_URN);

            final VirtualDevice virtualHumiditySensor =
                    gatewayDevice.createVirtualDevice(
                            humiditySensorEndpointId,
                            humidityDeviceModel);

            // Initialize the humidity sensor to the device model's maxThreshold default value
            defaultThreshold = virtualHumiditySensor.get("maxThreshold");
            humiditySensor.setMaxThreshold(defaultThreshold);

            /*
             * Set callbacks on the virtual devices to set attributes and
             * control the actual devices.
             */

            /*
             * For the temperatureSensor model, the min and max threshold values
             * can be written. Create a callback that will handle setting
             * the value on the temperature sensor device.
             */
            virtualTemperatureSensor.setOnChange(
                new VirtualDevice.ChangeCallback<VirtualDevice>() {
                    public void onChange(
                        VirtualDevice.ChangeEvent<VirtualDevice> event) {

                        VirtualDevice virtualDevice =
                            event.getVirtualDevice();
                        VirtualDevice.NamedValue<?> namedValues =
                            event.getNamedValue();
                        StringBuilder msg =
                            new StringBuilder(new Date().toString())
                                    .append(" : ")
                                    .append(virtualDevice.getEndpointId())
                                    .append(" : onChange : ");

                        boolean first = true;
                        for (VirtualDevice.NamedValue<?> namedValue =
                             namedValues;
                             namedValue != null;
                             namedValue = namedValue.next()) {
                            final String attribute = namedValue.getName();
                            final Object value = namedValue.getValue();
                            if (!first) {
                                msg.append(',');
                            } else {
                                first = false;
                            }

                            if (MIN_THRESHOLD_ATTRIBUTE.equals(attribute)) {
                                int min = ((Integer)value).intValue();
                                msg.append("\"minThreshold\"=");
                                msg.append(min);
                                if (temperatureSensor.isPoweredOn()) {
                                    // Update the actual device
                                    temperatureSensor.setMinThreshold(min);
                                }
                            } else if (MAX_THRESHOLD_ATTRIBUTE.equals(
                                    attribute)) {
                                int max = ((Integer)value).intValue();
                                msg.append("\"maxThreshold\"=");
                                msg.append(max);
                                if (temperatureSensor.isPoweredOn()) {
                                    // Update the actual device
                                    temperatureSensor.setMaxThreshold(max);
                                }
                            }

                            /*
                             * After this callback returns without an
                             * exception, the virtual device attributes
                             * will be updated locally and on the server.
                             */
                        }
                        if (!temperatureSensor.isPoweredOn()) {
                            msg.append(" ignored. Device is powered off");
                        }
                        display(msg.toString());
                    }
                });

            /*
             * The temperatureSensor model has a 'reset' action, which
             * resets the temperature sensor to factory defaults. Create
             * a callback that will handle calling reset on the temperature
             * sensor device.
             */
            virtualTemperatureSensor.setCallable("reset",
                new VirtualDevice.Callable<Void>() {
                    public void call(VirtualDevice virtualDevice,
                        Void not_used) {
                        StringBuilder msg =
                                new StringBuilder(new Date().toString())
                                        .append(" : ")
                                        .append(virtualDevice.getEndpointId())
                                        .append(" : Call : reset");
                        if (temperatureSensor.isPoweredOn()) {
                            temperatureSensor.reset();
                        } else {
                            msg.append(" ignored. Device is powered off");
                        }
                        display(msg.toString());
                        /*
                         * After the sensor is reset, next poll of the
                         * sensor will yield new min and max temp values
                         * and update the values in the VirtualDevice.
                         */
                    }
                });

            /*
             * The temperatureSensor model has a 'power' action, which
             * takes a boolean argument: true to power on the simulated device,
             * false to power off the simulated device. Create a callback that
             * will handle calling power on and power off on the temperature
             * sensor device.
             */
            virtualTemperatureSensor.setCallable("power",
                new VirtualDevice.Callable<Boolean>() {
                public void call(VirtualDevice virtualDevice, Boolean on) {
                    display(new Date().toString() + " : " +
                        virtualDevice.getEndpointId() +
                        " : Call : \"power\"=" + on);
                    boolean isOn = temperatureSensor.isPoweredOn();
                    if (on != isOn) {
                        temperatureSensor.power(on);
                    }

                    /*
                     * Avoid re-entering the library, let the polling
                     * loop handle the attribute update.
                     */
                }
            });

            /*
             * For the humiditySensor model, the maxThreshold attribute can be
             * written. Create a callback for setting the maxThreshold on the
             * humidity sensor device.
             */
            virtualHumiditySensor.setOnChange("maxThreshold",
                new VirtualDevice.ChangeCallback<VirtualDevice>() {
                    public void onChange(
                            VirtualDevice.ChangeEvent<VirtualDevice> event) {
                        VirtualDevice virtualDevice = event.getVirtualDevice();
                        VirtualDevice.NamedValue<?> namedValue =
                            event.getNamedValue();
                        Integer value = (Integer)namedValue.getValue();
                        display(new Date().toString() + " : " +
                            virtualDevice.getEndpointId() +
                            " : onChange : \"maxThreshold\"=" + value);

                        humiditySensor.setMaxThreshold(value);
                    }
                });


            /*
             * Create a handler for errors that may be generated when trying
             * to set values on the virtual device. The same callback is used
             * for both virtual devices.
             */
            final VirtualDevice.ErrorCallback<VirtualDevice> errorCallback =
                new VirtualDevice.ErrorCallback<VirtualDevice>() {
                    public void onError(VirtualDevice.ErrorEvent<VirtualDevice> event) {
                        VirtualDevice device = event.getVirtualDevice();
                        display(new Date().toString() + " : onError : " +
                            device.getEndpointId() +
                            " : \"" + event.getMessage() + "\"");
                    }
                };

            virtualTemperatureSensor.setOnError(errorCallback);
            virtualHumiditySensor.setOnError(errorCallback);

            /*
             * Create an Alert for max/min temperature threshold and another
             * for maximum humidity threshold. Alerts do not have to be
             * created new each time an alerting event occurs.
             */
            final Alert tooHotAlert = virtualTemperatureSensor
                .createAlert(TOO_HOT_ALERT);

            final Alert tooColdAlert = virtualTemperatureSensor
                .createAlert(TOO_COLD_ALERT);

            final Alert tooHumidAlert = virtualHumiditySensor
                .createAlert(TOO_HUMID_ALERT);

            display("\nCreated virtual humidity sensor " + humiditySensorEndpointId);
            display("Created virtual temperature sensor " + temperatureSensorEndpointId);
            display("\n\tPress enter to exit.\n");

            // Flags to ensure an alert is only generated when the threshold is crossed.
            boolean humidityAlerted = false;
            boolean tempAlerted = false;

            // Flag to know whether the temperature sensor was off and is now on.
            boolean wasOff = true;

            // The previous minimum temperature.
            // If the new minimum temperature is less than the previous,
            // update "minTemp" in the virtual TemperatureSensor
            double prevMinTemp = Double.MAX_VALUE;

            // The previous maximum temperature.
            // If the new maximum temperature is greater than the previous,
            // update "maxTemp" in the virtual TemperatureSensor
            double prevMaxTemp = Double.MIN_VALUE;

            // The previous minimum temperature threshold.
            // If the new minimum temperature threshold is not equal to the previous,
            // update "minThreshold" in the virtual TemperatureSensor
            int prevMinTempThreshold = temperatureSensor.getMinThreshold();

            // The previous maximum temperature threshold.
            // If the new maximum temperature threshold is not equal to the previous,
            // update "maxThreshold" in the virtual TemperatureSensor
            int prevMaxTempThreshold = temperatureSensor.getMaxThreshold();

            // The previous maximum humidity threshold.
            // If the new maximum humidity threshold is not equal to the previous,
            // update "maxThreshold" in the virtual HumiditySensor
            int prevMaxHumidityThreshold = humiditySensor.getMaxThreshold();

            mainLoop:
            for (; ; ) {

                int humidity = humiditySensor.getHumidity();

                StringBuilder consoleMessage = new
                        StringBuilder(new Date().toString());
                consoleMessage.append(" : ");
                consoleMessage.append(humiditySensorEndpointId);
                consoleMessage.append(" : Set : \"humidity\"=");
                consoleMessage.append(humidity);

                // Set the virtualilzed humidity sensor value in an 'update'. If
                // the humidity maxThreshold has changed, that maxThreshold attribute
                // will be added to the update (see the block below). A call to
                // 'finish' will commit the update.
                virtualHumiditySensor.update()
                        .set(HUMIDITY_ATTRIBUTE, humidity);

                // Send humidityThreshold if it has changed or is not the default
                int humidityThreshold = humiditySensor.getMaxThreshold();
                if (prevMaxHumidityThreshold != humidityThreshold) {
                    prevMaxHumidityThreshold = humidityThreshold;

                    consoleMessage
                            .append(",\"")
                            .append(MAX_THRESHOLD_ATTRIBUTE)
                            .append("\"=")
                            .append(humidityThreshold);

                    virtualHumiditySensor.set(MAX_THRESHOLD_ATTRIBUTE, humidityThreshold);
                }

                display(consoleMessage.toString());

                // Commit the update.
                virtualHumiditySensor.finish();

                if (humidity > humidityThreshold) {
                    if (!humidityAlerted) {
                        humidityAlerted = true;
                        consoleMessage = new StringBuilder(new Date().toString())
                                .append(" : ")
                                .append(virtualHumiditySensor.getEndpointId())
                                .append(" : Alert : \"")
                                .append(HUMIDITY_ATTRIBUTE)
                                .append("\"=")
                                .append(humidity)
                                .append(",\"")
                                .append(MAX_THRESHOLD_ATTRIBUTE)
                                .append("\"=")
                                .append(humidityThreshold);
                        display(consoleMessage.toString());
                        tooHumidAlert.set("humidity", humidity).raise();
                    }
                } else {
                    humidityAlerted = false;
                }

                if (temperatureSensor.isPoweredOn()) {

                    //
                    // Update the virtual device to reflect the new values.
                    //
                    final double temperature =
                            temperatureSensor.getTemp();
                    final String unit =
                            temperatureSensor.getUnit();
                    final double minTemp =
                            temperatureSensor.getMinTemp();
                    final double maxTemp =
                            temperatureSensor.getMaxTemp();
                    final int minThreshold =
                            temperatureSensor.getMinThreshold();
                    final int maxThreshold =
                            temperatureSensor.getMaxThreshold();

                    consoleMessage = new StringBuilder(new Date().toString())
                            .append(" : ")
                            .append(virtualTemperatureSensor.getEndpointId())
                            .append(" : Set : ")
                            .append("\"temp\"=" + temperature);

                    virtualTemperatureSensor.update()
                            .set("temp", temperature);

                    // There is no need to set startTime or units if the value has not changed.
                    if (wasOff) {
                        wasOff = false;
                        tempAlerted = false;

                        Date restartTime =
                                temperatureSensor.getStartTime();
                        consoleMessage
                                .append(",\"unit\"=").append(unit)
                                .append(",\"power\"=true")
                                .append(",\"startTime\"=").append(restartTime);
                        virtualTemperatureSensor
                                .set("startTime", restartTime)
                                .set("unit", unit);
                    }

                    // There is no need to set minTemp if the value has not changed.
                    if (minTemp != prevMinTemp) {
                        prevMinTemp = minTemp;
                        consoleMessage.append(",\"minTemp\"=").append(minTemp);
                        virtualTemperatureSensor.set("minTemp", minTemp);
                    }

                    // There is no need to set maxTemp if the value has not changed.
                    if (maxTemp != prevMaxTemp) {
                        prevMaxTemp = maxTemp;
                        consoleMessage.append(",\"maxTemp\"=").append(maxTemp);
                        virtualTemperatureSensor.set("maxTemp", maxTemp);
                    }

                    // There is no need to set minThreshold if the value has not changed, or is the default.
                    if (minThreshold != prevMinTempThreshold) {
                        prevMinTempThreshold = minThreshold;
                        consoleMessage.append(",\"minThreshold\"=").append(minThreshold);
                        virtualTemperatureSensor.set("minThreshold", minThreshold);
                    }

                    // There is no need to set maxThreshold if the value has not changed, or is the default.
                    if (maxThreshold != prevMaxTempThreshold) {
                        prevMaxTempThreshold = maxThreshold;
                        consoleMessage.append(",\"maxThreshold\"=").append(maxThreshold);
                        virtualTemperatureSensor.set("maxThreshold", maxThreshold);
                    }

                    display(consoleMessage.toString());

                    // finish() commits the update.
                    virtualTemperatureSensor.finish();

                    if (temperature > maxThreshold) {
                        if (!tempAlerted) {
                            tempAlerted = true;

                            consoleMessage = new StringBuilder(new Date().toString())
                                    .append(" : ")
                                    .append(virtualTemperatureSensor.getEndpointId())
                                    .append(" : Alert : ")
                                    .append("\"temp\"=" + temperature)
                                    .append(",\"unit\"=" + unit)
                                    .append(",\"maxThreshold\"=")
                                    .append(maxThreshold);

                            display(consoleMessage.toString());

                            tooHotAlert
                                    .set("temp", temperature)
                                    .set("unit", unit)
                                    .set("maxThreshold", maxThreshold)
                                    .raise();
                        }
                    } else if (temperature < minThreshold) {
                        if (!tempAlerted) {
                            tempAlerted = true;
                            consoleMessage = new StringBuilder(new Date().toString())
                                    .append(" : ")
                                    .append(virtualTemperatureSensor.getEndpointId())
                                    .append(" : Alert : ")
                                    .append("\"temp\"=" + temperature)
                                    .append(",\"unit\"=" + unit)
                                    .append(",\"minThreshold\"=")
                                    .append(minThreshold);

                            display(consoleMessage.toString());

                            tooColdAlert
                                    .set("temp", temperature)
                                    .set("unit", unit)
                                    .set("minThreshold", minThreshold)
                                    .raise();
                        }
                    } else {
                        tempAlerted = false;
                    }

                } else {
                    wasOff = true;
                }

                //
                // Wait SENSOR_POLLING_INTERVAL seconds before sending the next reading.
                // The SENSOR_POLLING_INTERVAL is broken into a number of smaller increments
                // to make the application more responsive to a keypress, which will cause
                // the application to exit.
                //
                for (int i = 0; i < number_of_loops; i++) {
                    Thread.sleep(sleepTime);

                    // when running under framework, use platform specific exit
                    if (!isUnderFramework) {
                        // User pressed the enter key while sleeping, exit.
                        if (System.in.available() > 0) {
                            break mainLoop;
                        }
                    } else if (exiting) {
                        break mainLoop;
                    }
                }

            }
        } catch (Throwable e) {
            // catching Throwable, not Exception:
            // could be java.lang.NoClassDefFoundError
            // which is not Exception

            displayException(e);
            if (isUnderFramework) throw new RuntimeException(e);
        } finally {
            // Dispose of the device client
            try {
                if (gatewayDevice != null) gatewayDevice.close();
            } catch (IOException ignored) {
            }
        }
    }

    private static void showUsage() {
        Class<?> thisClass = new Object() { }.getClass().getEnclosingClass();
        display("Usage: \n"
                + "java " + thisClass.getName()
                + " <trusted assets file> <trusted assets password> " +
                " [<temperature sensor activation id> [<humidity sensor activation id>]]\n");
    }

    private static void display(String s) {
        System.out.println(s);
    }
    private static void displayException(Throwable e) {
        StringBuffer sb = new StringBuffer(e.getMessage() == null ? 
                  e.getClass().getName() : e.getMessage());
        if (e.getCause() != null) {
            sb.append(".\n\tCaused by: ");
            sb.append(e.getCause());
        }
        System.out.println('\n' + sb.toString() + '\n');
        showUsage();
    }
}
